#include <sys/types.h>
#include <sys/socket.h>

#include <unistd.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * Note that you can not blindly assume that the below SYMBOL_ and
 * FUNCTION_ names will be correct on your system.
 *
 * 1) You will need to use a proggram that shows what systemcalls a binary
 * makes to verify that the symbolnames are correct; e.g. ktrace on BSD
 * systems, strace on Linux, truss on Solaris, etc.  Look at the output
 * and see what the name of the systemcall is.  E.g. socket(2) might
 * really be so_socket, read(2) might be read64(2), etc.
 *
 * 2) You will need to find out whether the symbols start with a 
 * an '_' or not.
 * 
 * Only after you have done both 1) and 2) can you be reasonably
 * sure the results are correct.  It does not gurantee for sure 
 * that the same thing will happen if LD_PRELOAD is used, but
 * most likly it will
 *
 * -- Michael Shuldman
 */

#if __linux__
#define LIBC                      "/lib/libc-2.3.6.so"

#define LIBRARY_SOCKET           LIBC
#define SYMBOL_SOCKET            "socket"
#define FUNCTION_SOCKET          socket

#define LIBRARY_READ             LIBC
#define SYMBOL_READ              "read"
#define FUNCTION_READ            read

#define LIBRARY_WRITE            LIBC
#define SYMBOL_WRITE             "write"
#define FUNCTION_WRITE           write

#elif __OpenBSD__ || __NetBSD__ 
#define LIBC                      "/usr/lib/libc.so"

#define LIBRARY_SOCKET           LIBC
#define SYMBOL_SOCKET            "socket"
#define FUNCTION_SOCKET          socket

#define LIBRARY_READ             LIBC
#define SYMBOL_READ              "read"
#define FUNCTION_READ            read

#define LIBRARY_WRITE            LIBC
#define SYMBOL_WRITE             "write"
#define FUNCTION_WRITE           write

#elif __FreeBSD__
#define LIBC                      "/usr/lib/libc.so"

#define LIBRARY_SOCKET           LIBC
#define SYMBOL_SOCKET            "_socket"
#define FUNCTION_SOCKET          _socket

#define LIBRARY_READ             LIBC
#define SYMBOL_READ              "_read"
#define FUNCTION_READ            _read

#define LIBRARY_WRITE            LIBC
#define SYMBOL_WRITE             "_write"
#define FUNCTION_WRITE           _write

#elif sun /* ! __FreeBSD__ */
#define LIBC                      "/usr/lib/libc.so"

#define LIBRARY_SOCKET           "/usr/lib/libsocket.so"
#define SYMBOL_SOCKET            "_so_socket"
#define FUNCTION_SOCKET          _so_socket

#define LIBRARY_READ             LIBC
#define SYMBOL_READ              "_read"
#define FUNCTION_READ            _read

#define LIBRARY_WRITE            LIBC
#define SYMBOL_WRITE             "_write"
#define FUNCTION_WRITE           _write

#elif __APPLE__ /* !SOLARIS */
#define LIBC                      "/usr/lib/libc.dylib"

#define LIBRARY_SOCKET           LIBC
#define SYMBOL_SOCKET            "socket"
#define FUNCTION_SOCKET          socket

#define LIBRARY_READ             LIBC
#define SYMBOL_READ              "read"
#define FUNCTION_READ            read

#define LIBRARY_WRITE            LIBC
#define SYMBOL_WRITE             "write"
#define FUNCTION_WRITE           write

#error "define platform"
#endif

static void *symbolfunction(const char *symbol, const char *library);

int socket_called;
int read_called;
int write_called;

int
FUNCTION_SOCKET(domain, type, protocol)
   int domain;
   int type;
   int protocol;
{
   typedef int (*SOCKET_FUNC_T)(int, int, int);
   SOCKET_FUNC_T function;
   
   socket_called = 1;
   function = (SOCKET_FUNC_T)symbolfunction(SYMBOL_SOCKET, LIBRARY_SOCKET);
   return function(domain, type, protocol);
}

ssize_t
FUNCTION_READ(d, buf, nbytes)
   int d;
   void *buf;
   size_t nbytes;
{
   typedef ssize_t (*READ_FUNC_T)(int, void *, size_t);
   READ_FUNC_T function;
   
   read_called = 1;
   function = (READ_FUNC_T)symbolfunction(SYMBOL_READ, LIBRARY_READ);
   return function(d, buf, nbytes);
}

ssize_t
FUNCTION_WRITE(d, buf, nbytes)
   int d;
   const void *buf;
   size_t nbytes;
{
   typedef ssize_t (*WRITE_FUNC_T)(int, const void *, size_t);
   WRITE_FUNC_T function;
   
   write_called = 1;
   function = (WRITE_FUNC_T)symbolfunction(SYMBOL_WRITE, LIBRARY_WRITE);
   return function(d, buf, nbytes);
}


static void *
symbolfunction(symbol, library)
   const char *symbol;
   const char *library;
{
   void *handle, *function;

   if ((handle = dlopen(library, RTLD_LAZY)) == NULL) {
      fprintf(stderr, "dlopen(%s): %s\n", library, dlerror());
      exit(1);
   }

   if ((function = dlsym(handle, symbol)) == NULL) {
      fprintf(stderr, "dlsym(%s): %s\n", symbol, dlerror());
      exit(1);
   }

   return function;
}
